unit snd_main;
interface

uses
  Windows;
  
const
  SND_MAX_CHANNELS = 64;    // Nombre maximum de canaux
  SND_MAX_TIME     = 3000;  // Maximum life time after end of channel playing (ms)

var
  snd_count_playing : Integer = 0;     // on playing channels count
  snd_count_active  : Integer = 0;     // active channels count (buffer <> nil)
  snd_count_samples : Integer = 0;     // samples count
  snd_volume  : Integer = 100;   // Default max sample volume
  snd_freq    : Integer = 22050; // Default frequency
  snd_off     : Boolean = False; // Sound off
// pan control
  snd_swap           : Boolean = False; // Swap Left-Right channels
  snd_factor_pan     : Integer = 8;
  snd_factor_rolloff : Integer = 2048;

  function SND_init(Handle: HWND): Boolean;
  procedure SND_free;
  function SND_add(const FileName: string; Group: Integer = 0): Integer;
  function SND_del(SID: Integer): Boolean;
  function SND_play(SID: Integer; X: Single = 0; Y: Single = 0; Loop: Boolean = False): Integer;
  procedure SND_stop(CID: Integer);
  procedure SND_channel_pos(CID: Integer; X, Y: Single);
  procedure SND_global_pos(X, Y: Single);
  procedure SND_global_volume(Volume: Integer = 100);
  procedure SND_global_freq(Frequency: Integer = 22050);
  procedure SND_update;
  procedure SND_refresh_begin(Group: Integer);
  procedure SND_refresh_end(Group: Integer);

implementation

// DirectSound v5.0 interfaces & constants
const
  DS_OK                         = $00000000;
  DSSCL_PRIORITY                = $00000002;
  DSBPLAY_LOOPING               = $00000001;
  DSBPAN_LEFT                   = -10000;
  DSBPAN_RIGHT                  = 10000;
  DSBVOLUME_MIN                 = -10000;
  DSBVOLUME_MAX                 = 0;
  DSBSTATUS_PLAYING             = $00000001;
  DSBSTATUS_BUFFERLOST          = $00000002;
  DSBLOCK_ENTIREBUFFER          = $00000002;

  DSBCAPS_STATIC                = $00000002;
  DSBCAPS_LOCSOFTWARE           = $00000008;
  DSBCAPS_CTRLFREQUENCY         = $00000020;
  DSBCAPS_CTRLPAN               = $00000040;
  DSBCAPS_CTRLVOLUME            = $00000080;
  DSBCAPS_GLOBALFOCUS           = $00008000;

type
  TDSBufferDesc = packed record
    dwSize          : DWORD;
    dwFlags         : DWORD;
    dwBufferBytes   : DWORD;
    dwReserved      : DWORD;
    lpwfxFormat     : Pointer;
    guid3DAlgorithm : TGUID;
  end;

  IDirectSound = interface;
  IDirectSoundBuffer = interface;

//
// IDirectSound
//
  PIDirectSound = ^IDirectSound;
  IDirectSound = interface (IUnknown)
    ['{279AFA83-4981-11CE-A521-0020AF0BE560}']
    // IDirectSound methods
    function CreateSoundBuffer(const lpDSBufferDesc: TDSBufferDesc;
        out lpIDirectSoundBuffer: IDirectSoundBuffer;
        pUnkOuter: IUnknown) : HResult; stdcall;
    function GetCaps(lpDSCaps: Pointer) : HResult; stdcall;
    function DuplicateSoundBuffer(lpDsbOriginal: IDirectSoundBuffer;
        out lpDsbDuplicate: IDirectSoundBuffer) : HResult; stdcall;
    function SetCooperativeLevel(hwnd: HWND; dwLevel: DWORD) : HResult; stdcall;
    function Compact: HResult; stdcall;
    function GetSpeakerConfig(var lpdwSpeakerConfig: DWORD) : HResult; stdcall;
    function SetSpeakerConfig(dwSpeakerConfig: DWORD) : HResult; stdcall;
    function Initialize(lpGuid: PGUID) : HResult; stdcall;
  end;

//
// IDirectSoundBuffer
//
  IDirectSoundBuffer = interface (IUnknown)
    ['{279AFA85-4981-11CE-A521-0020AF0BE560}']
    // IDirectSoundBuffer methods
    function GetCaps(lpDSCaps: Pointer) : HResult; stdcall;
    function GetCurrentPosition
        (lpdwPlayPosition, lpdwReadPosition : PDWORD) : HResult; stdcall;
    function GetFormat(lpwfxFormat: Pointer; dwSizeAllocated: DWORD;
        lpdwSizeWritten: PDWORD) : HResult; stdcall;
    function GetVolume(var lplVolume: integer) : HResult; stdcall;
    function GetPan(var lplPan: integer) : HResult; stdcall;
    function GetFrequency(var lpdwFrequency: DWORD) : HResult; stdcall;
    function GetStatus(var lpdwStatus: DWORD) : HResult; stdcall;
    function Initialize(lpDirectSound: IDirectSound;
        const lpcDSBufferDesc: TDSBufferDesc) : HResult; stdcall;
    function Lock(dwWriteCursor, dwWriteBytes: DWORD;
        var lplpvAudioPtr1: Pointer; var lpdwAudioBytes1: DWORD;
        var lplpvAudioPtr2: Pointer; var lpdwAudioBytes2: DWORD;
        dwFlags: DWORD) : HResult; stdcall;
    function Play(dwReserved1,dwReserved2,dwFlags: DWORD) : HResult; stdcall;
    function SetCurrentPosition(dwPosition: DWORD) : HResult; stdcall;
    function SetFormat(lpcfxFormat: Pointer) : HResult; stdcall;
    function SetVolume(lVolume: integer) : HResult; stdcall;
    function SetPan(lPan: integer) : HResult; stdcall;
    function SetFrequency(dwFrequency: DWORD) : HResult; stdcall;
    function Stop: HResult; stdcall;
    function Unlock(lpvAudioPtr1: Pointer; dwAudioBytes1: DWORD;
        lpvAudioPtr2: Pointer; dwAudioBytes2: DWORD) : HResult; stdcall;
    function Restore: HResult; stdcall;
  end;

var
  DSoundDLL : HMODULE;
  DirectSoundCreate : function (lpGuid: PGUID; out ppDS: IDirectSound; pUnkOuter: IUnknown): HResult; stdcall;

type
  TChannel = object
    Buffer   : IDirectSoundBuffer; // Direct Sound buffer
    SID      : Integer;  // Source Sample ID
    Sample   : Boolean;  // is Source Sample?
    Playing  : Boolean;  // on Playing?
    Timer    : DWORD;    // Idle timer for channel (0 - SND_MAX_TIME see SND_update)
    Group    : Integer;  // Sound group for refresh
    Flag     : Boolean;  // Refresh flag (see SND_refresh_begin/end)
    Ref      : Integer;  // References count
    X, Y     : Single;   // Channel position
    FileName : string;   // Sample file name
  end;

var
  snd_ready : Boolean = False;
  snd_pos   : record
                X, Y : Single;
              end = (X: 0; Y: 0);
  DSMain    : IDirectSound;
  Channels  : array [0..SND_MAX_CHANNELS - 1] of TChannel;

function SND_GetFreeID: Integer;
var
  i : Integer;
begin
//== search free channel ("fully free" has high priority)
  Result := -1;
  for i := 0 to SND_MAX_CHANNELS - 1 do
    with Channels[i] do
      if (not Sample) and
         ((not Playing) or (Buffer = nil)) then
      begin
        Result := i;
        if Buffer = nil then
          break;
      end;
end;

procedure SND_SetVolume(CID, Volume: Integer);
begin
//== Set channel volume
  with Channels[CID] do
    Buffer.SetVolume(trunc(sqrt(Volume * snd_volume * 0.0001) * (DSBVOLUME_MAX - DSBVOLUME_MIN) + DSBVOLUME_MIN));
end;

procedure SND_Calc(CID: Integer);
var
  i    : Integer;
  dist : Single;
begin
//== Recalc channel volume and pan
  with Channels[CID] do
  begin
    i := trunc((X - snd_pos.X) * snd_factor_pan);
    if i < DSBPAN_LEFT  then i := DSBPAN_LEFT;
    if i > DSBPAN_RIGHT then i := DSBPAN_RIGHT;
    if snd_swap then
      i := -i;
    Buffer.SetPan(i);
    dist := sqrt(sqr(X - snd_pos.X) + sqr(Y - snd_pos.Y));
    if dist <= snd_factor_rolloff then
      i := trunc((1 - dist/snd_factor_rolloff) * 100)
    else
      i := 0;
    SND_SetVolume(CID, i);
 end;
end;

///////////////////////////////////////////////////
//== Sound initialization
function SND_init(Handle: HWND): Boolean;
var
  i : Integer;
begin
  Result := False;
  if (DSoundDLL = 0) or (@DirectSoundCreate = nil) then
    Exit;
  if DirectSoundCreate(nil, DSMain, nil) <> DS_OK then
    Exit;
  if DSMain.SetCooperativeLevel(Handle, DSSCL_PRIORITY) <> DS_OK then
    ;//Log...

// Initialize all channels
  for i := 0 to SND_MAX_CHANNELS - 1 do
    with Channels[i] do
    begin
      Buffer  := nil;
      Ref     := 0;
      Sample  := False;
      SID     := 0;
      Flag    := False;
      Playing := False;
    end;

  Result := True;
  snd_ready := True;
end;

//== Sound deinitialization
procedure SND_free;
var
  i : Integer;
begin
  if snd_ready then
  begin
    for i := 0 to SND_MAX_CHANNELS - 1 do
      Channels[i].Buffer := nil;
    DSMain     := nil;
    snd_ready  := False;
  end;
end;

//== Loading wav (PCM) file
function SND_add(const FileName: string; Group: Integer): Integer;
var
  i   : Integer;
  NID : Integer;

// Microsoft Wave file format header
  WaveHeader : record
    RIFF : array [0..3] of Char;
    zero : DWORD;
    WAVE : array [0..3] of Char;
    res  : array [0..1] of DWORD;
    wfmt : packed record
             wFormatTag      : Word;  { format type }
             nChannels       : Word;  { number of channels (i.e. mono, stereo, etc.) }
             nSamplesPerSec  : DWORD; { sample rate }
             nAvgBytesPerSec : DWORD; { for buffer estimation }
             nBlockAlign     : Word;  { block size of data }
             wBitsPerSample  : Word;  { number of bits per sample of mono data }
             cbSize          : Word;  { the count in bytes of the size of }
           end;
    w    : Word;
    Size : DWORD;
  end;

  F          : File;
  ap1, ap2   : Pointer;
  as1, as2   : DWORD;
  BufferDesc : TDSBufferDesc;
label
  ext;
begin
  Result := -1;
  if not snd_ready then
    Exit;
// if Sample is loaded
  for i := 0 to SND_MAX_CHANNELS - 1 do
    if (Channels[i].FileName = FileName) and
       (Channels[i].Buffer <> nil) then
    begin
    // use ready sample
      Result := i;
      inc(Channels[i].Ref);
      Exit;
    end;

  NID := SND_GetFreeID;
  if NID = -1 then
    Exit;

// Load sample from wav file
  FileMode := 64;
  AssignFile(F, FileName);
  {$I-}
  Reset(F, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    //Log('File not found')
    Exit;
  end;
  BlockRead(F, WaveHeader, SizeOf(WaveHeader));

// read in some properties of the wave data in the file
  with WaveHeader, wfmt do
  begin
    if (RIFF <> 'RIFF') or (WAVE <> 'WAVE') or
       (wBitsPerSample < 8) then
    begin
      //Log('invalid wave format');
      goto ext;
    end;
    nBlockAlign     := wBitsPerSample div 8 * nChannels;
    nAvgBytesPerSec := nSamplesPerSec * nBlockAlign;
  end;

// create static buffer description
  ZeroMemory(@BufferDesc, SizeOf(BufferDesc));
  with BufferDesc do
  begin
    dwSize  := SizeOf(BufferDesc);
    dwFlags := DSBCAPS_LOCSOFTWARE or 
               DSBCAPS_STATIC or
               DSBCAPS_GLOBALFOCUS or
               DSBCAPS_CTRLFREQUENCY or
               DSBCAPS_CTRLPAN or
               DSBCAPS_CTRLVOLUME;
    dwBufferBytes := WaveHeader.Size;
    lpwfxFormat   := @WaveHeader.wfmt;
  end;

// create channel sound buffer
  Channels[NID].Buffer := nil;
  if DSMain.CreateSoundBuffer(BufferDesc, Channels[NID].Buffer, nil) <> DS_OK then
  begin
    //Log('Sound Error: Can''t create primary sound buffer');
    goto ext;
  end;

// lock buffer data
  ap2 := nil;
  as2 := 0;
  if Channels[NID].Buffer.Lock(0, 0, ap1, as1, ap2, as2,
                               DSBLOCK_ENTIREBUFFER) <> DS_OK then
  begin
    //Log('Sound Error: Can''t lock buffer data');
    goto ext;
  end;

// write data to buffer
  BlockRead(F, ap1^, WaveHeader.Size);

// unlock buffer data
  if Channels[NID].Buffer.Unlock(ap1, as1, ap2, as2) <> DS_OK then
  begin
    //Log('Sound Error: Can''t unlock buffer data');
    goto ext;
  end;
  Channels[NID].FileName := FileName;
  Channels[NID].SID      := NID;
  Channels[NID].Ref      := 1;
  Channels[NID].Sample   := True;
  Channels[NID].Flag     := False;
  Channels[NID].Playing  := False;
  Channels[NID].Group    := Group;
  Result := NID;
ext:
  CloseFile(F);
end;

//== Delete sample from list
function SND_del(SID: Integer): Boolean;
begin
  Result := False;
  if not snd_ready then
    Exit;

  if (SID > -1) and (SID < SND_MAX_CHANNELS) and
     (Channels[SID].Buffer <> nil) then
  begin
    if Channels[SID].Ref = 1 then
      Channels[SID].Buffer := nil
    else
      dec(Channels[SID].Ref);
    Result := True;
  end;
end;

//== Play sound sample
function SND_play(SID: Integer; X, Y: Single; Loop: Boolean): Integer;
var
  i   : Integer;
  NID : Integer;  // New Channel ID
begin
  Result := -1;
  if (not snd_ready) or snd_off or
     (SID < 0) or (SID >= SND_MAX_CHANNELS) or
     (Channels[SID].Buffer = nil) or
     (not Channels[SID].Sample) then
    Exit;
    
  NID := -1;
// searching stopped channel with ID = SID
  for i := 0 to SND_MAX_CHANNELS - 1 do
    if (Channels[i].SID = SID)   and
       (not Channels[i].Playing) and
       (Channels[i].Buffer <> nil) then
    begin
    // and play
      NID := i;
      break;
    end;

// else duplicate sample buffer into free channel
  if NID = -1 then
  begin
    NID := SND_GetFreeID;
  // if no channels found - can't play sample
    if NID = -1 then
      Exit;
  // duplicate sample buffer to free channel (NID)
    Channels[NID].SID    := SID;
    Channels[NID].Flag   := False;
    Channels[NID].Sample := False;
    Channels[NID].Buffer := nil;
    DSMain.DuplicateSoundBuffer(Channels[SID].Buffer,
                                Channels[NID].Buffer);
  end;

  if not Channels[NID].Sample then
    Channels[NID].Ref := 1;
  Channels[NID].X := X;
  Channels[NID].Y := Y;
  with Channels[NID] do
  begin
    SND_Calc(NID);
    Buffer.SetFrequency(snd_freq);
    Playing := Buffer.Play(0, 0, DSBPLAY_LOOPING and Byte(Loop)) = DS_OK;
  end;
  Result := NID;
end;

//== Stop playing channel(s)
procedure SND_stop(CID: Integer);
var
  i : Integer;
begin
  if (not snd_ready) or
     (CID < -1) or (CID >= SND_MAX_CHANNELS) then
    Exit;

  if CID > -1 then
  begin
    with Channels[CID] do
      if Playing and (Buffer <> nil) then
        Buffer.Stop;
  end else // CID = -1 - stop all channels
    for i := 0 to SND_MAX_CHANNELS - 1 do
      with Channels[i] do
        if Playing and (Buffer <> nil) then
          Buffer.Stop;
end;

//== Set channel position
procedure SND_channel_pos(CID: Integer; X, Y: Single);
begin
  if (not snd_ready) or
     (CID < 0) or (CID >= SND_MAX_CHANNELS) or
     (Channels[CID].Buffer = nil) then
    Exit;
  Channels[CID].X := X;
  Channels[CID].Y := Y;
  SND_Calc(CID);
end;

//== Set listener position
procedure SND_global_pos(X, Y: Single);
begin
  if not snd_ready then
    Exit;

  snd_pos.X := X;
  snd_pos.Y := Y;
end;

//== Set sound volume
procedure SND_global_volume(Volume: Integer = 100);
var
  i   : Integer;
begin
  if not snd_ready then
    Exit;

  snd_volume := Volume;
  if snd_volume > 100 then snd_volume := 100;
  if snd_volume <   0 then snd_volume := 0;
  for i := 0 to SND_MAX_CHANNELS - 1 do
    with Channels[i] do
      if Playing and (Buffer <> nil) then
        SND_SetVolume(i, Volume);
end;

//== Set sound frequency
procedure SND_global_freq(Frequency: Integer = 22050);
var
  i : Integer;
begin
  if not snd_ready then
    Exit;
  snd_freq := Frequency;
  for i := 0 to SND_MAX_CHANNELS - 1 do
    with Channels[i] do
      if Playing and (Buffer <> nil) then
        Buffer.SetFrequency(snd_freq);
end;

//== Refresh before reloading
procedure SND_refresh_begin(Group: Integer);
var
  i : Integer;
begin
//== -1 - refresh all samples
  for i := 0 to SND_MAX_CHANNELS - 1 do
    if (Group and Channels[i].Group > 0) or (group = -1) then
      Channels[i].Flag := True;
end;

//== End reloading
procedure SND_refresh_end(Group: Integer);
var
  i : Integer;
begin
  for i := 0 to SND_MAX_CHANNELS - 1 do
    if ((Group and Channels[i].Group > 0) or (Group = -1)) and Channels[i].Flag then
      SND_del(i);
end;

//== Updating channels and statistics
procedure SND_update;
var
  i       : Integer;
  Status  : DWORD;
  CurTime : DWORD;
begin
  if not snd_ready then
    Exit;

  snd_count_playing := 0;
  snd_count_active  := 0;
  snd_count_samples := 0;
  CurTime := GetTickCount;
// Recalc channels status
  for i := 0 to SND_MAX_CHANNELS - 1 do
    with Channels[i] do
      if Buffer <> nil then
      begin
        Buffer.GetStatus(Status);
        if Status and DSBSTATUS_BUFFERLOST <> 0 then
          Buffer.Restore
        else
          if status and DSBSTATUS_PLAYING <> 0 then
          begin
            SND_Calc(i);
            inc(snd_count_playing);
          end else
          begin
            if not Sample then
              if Playing then
                Timer := CurTime
              else
                if CurTime - Timer > SND_MAX_TIME then
                begin
                  Buffer := nil;
                  continue;
                end;
            Playing := False;
          end;
        if Sample then
          inc(snd_count_samples);
        inc(snd_count_active);
      end;
end;

initialization
  DSoundDLL := LoadLibrary('DSound.dll');
  DirectSoundCreate := GetProcAddress(DSoundDLL, 'DirectSoundCreate');

finalization
  if DSoundDLL <> 0 then FreeLibrary(DSoundDLL);

end.
